"""Utils to parse JS"""

import os
from pathlib import Path
from functools import wraps
from itertools import chain
from typing import Optional
import json

import esprima

AstScript = esprima.nodes.Script
AstNode = esprima.nodes.Node


# TODO: Parse out js docs and add them to py mirror function


class UnknownNodeType(ValueError):
    """To be raised when an AST node type wasn't handled by a function"""

    def __init__(self, node_type, context=''):
        super().__init__(f'Unknown type {context}: {node_type}')


def if_none_output_raise_unknown_type(context=''):
    def _if_none_output_raise_wrapper(func):
        @wraps(func)
        def _func(*args, **kwargs):
            output = func(*args, **kwargs)
            if output is None:
                # Note we only extract from args, if in kwargs won't work
                # For that reason, we should enforce the node to be the first,
                # positional only, argument in func
                node, *_ = args
                raise UnknownNodeType(node.type, context)
            else:
                return output

        return _func

    return _if_none_output_raise_wrapper


def parse_js_code(js_code: str, encoding: Optional[str] = None) -> AstScript:
    if os.path.isfile(js_code):
        js_code = Path(js_code).read_text(encoding=encoding)
    return esprima.parse(js_code)


def extract_function_def_parts(func_def_code: str) -> str:
    """Extract the body of a function definition from its code"""
    body_tree = parse_js_code(func_def_code).body[0].body

    # Find the position of the first opening brace
    start = func_def_code.find('{', body_tree.start)
    # Find the position of the last closing brace
    end = func_def_code.rfind('}', body_tree.start, body_tree.end)

    return (
        func_def_code[:start],
        func_def_code[start + 1 : end],
        func_def_code[end + 1 :],
    )


def append_to_func_body(func_def_code: str, code_to_append: str) -> str:
    """Append code to the body of a function definition"""
    pre, body, post = extract_function_def_parts(func_def_code)
    return pre + '{' + body + code_to_append + '}' + post


@if_none_output_raise_unknown_type('to extract obj name from')
def _extract_obj_name(x: AstNode, /):
    if x.type == 'Identifier':
        return x.name
    elif x.type == 'MemberExpression':
        object_name = _extract_obj_name(x.object)
        property_name = _extract_obj_name(x.property)
        return f'{object_name}.{property_name}'


@if_none_output_raise_unknown_type('to extract params from')
def _extract_params(x: AstNode, /):
    if x.type == 'Identifier':
        return dict(name=x.name)
    elif x.type == 'AssignmentPattern':
        return dict(name=x.left.name, default=x.right.value)


def extract_js_func_params(params_list):
    return map(_extract_params, params_list)


def _extract_params_from_function_expression(x):
    return extract_js_func_params(x.params)


def extract_func_name_and_params(ast_node: AstNode):
    """Extract one or several function ``(name, params)`` pair(s) from an AST node"""
    x = ast_node
    if x.type == 'FunctionDeclaration':
        yield x.id.name, list(extract_js_func_params(x.params))
    elif x.type == 'AssignmentExpression':
        yield (
            _extract_obj_name(x.left),
            list(extract_js_func_params(x.right.params)),
        )
    elif x.type == 'VariableDeclarator':
        yield x.id.name, list(extract_js_func_params(x.init.params))
    elif x.type == 'VariableDeclaration':
        # Here we may have several declarations, so we use yield from
        yield from chain.from_iterable(
            map(extract_func_name_and_params, x.declarations)
        )
    elif x.type == 'ExpressionStatement':
        yield from extract_func_name_and_params(x.expression)


def func_name_and_params_pairs(js_code: str, *, encoding=None):
    """
    Get ``(name, params)`` pairs of function definitions extracted from ``js_code``.

    >>> js_code = '''
    ... function foo(a, b="hello", c= 3) {
    ...     return a + b.length * c
    ... }
    ... const bar = (y, z = 1) => y * z
    ... func.assigned.to.nested.prop = function (x) {
    ...     return x + 3
    ... }'''
    >>> assert list(func_name_and_params_pairs(js_code)) == [
    ...     ('foo', [
    ...         {'name': 'a'},
    ...         {'name': 'b', 'default': 'hello'},
    ...         {'name': 'c', 'default': 3}
    ...     ]),
    ...     ('bar', [
    ...         {'name': 'y'},
    ...         {'name': 'z', 'default': 1}
    ...     ]),
    ...     ('func.assigned.to.nested.prop', [
    ...         {'name': 'x'}
    ...     ])
    ... ]

    """
    ast_script = parse_js_code(js_code, encoding=encoding)
    for ast_node in ast_script.body:
        # TODO: Check for type of ast_node before calling extract_func_name_and_params
        #  For now, just ignoring errors:
        try:
            yield from extract_func_name_and_params(ast_node)
        except Exception as e:
            print(f'Exception while parsing {ast_node}: {e}')


def dflt_py_to_js_value_trans(x):
    if isinstance(x, bool):
        return str(x).lower()
    elif isinstance(x, str):
        return f'"{x}"'  # surround with quotes
    elif isinstance(x, dict):
        return json.dumps(x)
    return x


import re
import uuid
from typing import Tuple

_dflt_patterns_for_html_ids = (
    r'id="([^"]+)"',  # HTML id attributes
    r'#([a-zA-Z][\w\-]*)',  # CSS IDs
    r'document\.getElementById\("([^"]+)"\)',  # JS getElementById
)


def replace_ids_in_code(
    code: str, patterns: Tuple[str] = _dflt_patterns_for_html_ids
) -> (str, dict):
    """
    Replaces IDs in a mixed HTML, CSS, and JS string with unique versions.

    Args:
    - code (str): The input code containing HTML, CSS, and/or JS.
    - patterns (tuple): A tuple of regex patterns to match IDs.

    Returns:
    - tuple: A tuple containing the modified code and a dictionary of old-to-new ID mappings.

    This function is useful to make id-unique copies of HTML/JS components that have IDs.
    This need was noticed and discussed in: https://github.com/i2mint/jy/discussions/3#discussioncomment-6862843

    """
    # Patterns to match IDs
    patterns = [
        r'id="([^"]+)"',  # HTML id attributes
        r'#([a-zA-Z][\w\-]*)',  # CSS IDs
        r'document\.getElementById\("([^"]+)"\)',  # JS getElementById
    ]

    replaced = {}

    # Generate a unique ID
    def unique_id(match):
        old_id = match.group(1)
        new_id = old_id + '_' + str(uuid.uuid4()).replace('-', '')[:8]
        replaced[old_id] = new_id
        return match.group(0).replace(old_id, new_id)

    # Replace all IDs in the code
    for pattern in patterns:
        code = re.sub(pattern, unique_id, code)

    return code, replaced



"""Python portals to JS"""

from typing import Any, Optional
from functools import partial

from dol.signatures import Sig

from jy.js_parse import dflt_py_to_js_value_trans, func_name_and_params_pairs


# TODO: Add value transformer (routing). For example, booleans, True->true
def _js_func_call(
    *args,
    __sig,
    __func_call_template,
    __value_trans=dflt_py_to_js_value_trans,
    __apply_defaults=True,
    **kwargs,
):
    _kwargs = __sig.map_arguments(args, kwargs, apply_defaults=__apply_defaults)
    inputs = map(__value_trans, _kwargs.values())
    inputs = ', '.join(map(str, inputs))
    return __func_call_template.format(inputs=inputs)


# TODO: value_trans only sees value. Might want to transform according to other
#  contextural information, like the parameter name, or the function name, etc.
def mk_py_binder_func(
    name,
    params,
    *,
    prefix='',
    suffix='',
    doc='',
    value_trans=dflt_py_to_js_value_trans,
    apply_defaults=True,
):
    *_, func_name = name.split('.')  # e.g. object.containing.func --> func
    func_call_template = prefix + name + '({inputs})' + suffix

    sig = Sig.from_params(params)
    js_func_call = partial(
        _js_func_call,
        __sig=sig,
        __func_call_template=func_call_template,
        __value_trans=value_trans,
        __apply_defaults=apply_defaults,
    )
    js_func_call = sig(js_func_call)
    js_func_call.__name__ = func_name
    js_func_call.__doc__ = doc

    return js_func_call


class JsBridge:
    """
    The default class to make instances that will contain methods that mirror JS
    function calls
    """


def add_js_funcs(
    js_code: str,
    *,
    obj: Any = None,
    name: Optional[str] = None,
    encoding: Optional[str] = None,
    forbidden_method_names=(),
    apply_defaults=True,
    value_trans=dflt_py_to_js_value_trans,
):
    '''
    Add js call functions as attributes to an object.

    If object is not given, ``add_js_funcs`` will use a new ``JsBridge`` instance.

    >>> js_code = """
    ... function foo(a, b="hello", c= 3) {
    ...     return a + b.length * c
    ... }
    ... const bar = (y, z = 1) => y * z
    ... func.assigned.to.nested.prop = function (x) {
    ...     return x + 3
    ... }
    ... """
    >>> js = add_js_funcs(js_code)
    >>> from inspect import signature
    >>> list(vars(js))
    ['foo', 'bar', 'prop']
    >>> signature(js.foo)
    <Sig (a, b='hello', c=3)>
    >>> js.foo(1, 'hi')
    'foo(1, "hi", 3)'
    >>> js.prop('up')
    'func.assigned.to.nested.prop("up")'
    '''

    if obj is None:
        obj = JsBridge()

    forbidden_method_names = set(forbidden_method_names)

    for full_func_ref, params in func_name_and_params_pairs(js_code, encoding=encoding):
        *_, func_name = full_func_ref.split('.')  # e.g. object.containing.func --> func
        # TODO: Could specify a recovery function that finds another name instead of
        #  raising an error
        if func_name in forbidden_method_names:
            raise ValueError(
                f'This func name was already used, or mentioned in '
                f'the forbidden_method_names argument: {func_name}'
            )
        setattr(
            obj,
            func_name,
            mk_py_binder_func(
                full_func_ref,
                params,
                value_trans=value_trans,
                apply_defaults=apply_defaults,
            ),
        )

    if name:
        obj.__name__ = name

    return obj



"""
In order to solve this problem: 
https://github.com/i2mint/jy/discussions/3#discussioncomment-6862843
I wrote some code to find and replace ids in html, css, and js strings.

It seems to complex and not very robust. I'm sure there are better ways to do this.
I'm still posting this here in case I want to resurect it.

This code has tests!

"""
# from jy.js_parse import replace_tokens, string_kind, replace_ids, default_token_replacer

import re
from functools import partial
from typing import Callable, Optional, Union


def replace_tokens(string: str, token_extractor: Callable, token_replacer: Callable):
    """
    Replace tokens in a string based on patterns provided in the token_extractor.
    
    Parameters:
    - string: The input string in which tokens need to be replaced.
    - token_extractor: A dictionary mapping token classes to their regex patterns.
    - token_replacer: A function that determines how a matched token should be replaced.
    
    Returns:
    - new_string: The string with tokens replaced.
    - replaced: A dictionary mapping old tokens to their replacements.
    
    >>> new_string, replaced = replace_tokens(
    ...     'Hello, my id="name" is John.', 
    ...     {"id": r'id="([^"]+)"'}, 
    ...     default_token_replacer
    ... )
    >>> list(replaced)
    ['id="name"']
    >>> replaced['id="name"'].startswith('id="name_')
    True
    >>> replaced['id="name"']  # doctest: +SKIP
    'id="name_db72fa7f"'  # random suffix was added
    """

    replaced = {}
    new_string = string

    for token_class, token_pattern in token_extractor.items():
        for match in re.finditer(token_pattern, new_string):
            old_token = match.group(0)
            new_token = token_replacer(old_token, token_class)
            replaced[old_token] = new_token
            new_string = new_string.replace(old_token, new_token)

    return new_string, replaced


import uuid


def default_token_replacer(matched_token: str, token_class: str):
    """
    Default token replacer that appends a unique suffix to the matched token.
    
    Parameters:
    - matched_token: The token that was matched.
    - token_class: The class of the token (e.g., "id").
    
    Returns:
    - The replaced token with a unique suffix.
    
    >>> default_token_replacer('id="test"', 'id')
    'id="test_<some_unique_suffix>"'
    """
    unique_suffix = str(uuid.uuid4()).replace('-', '')[
        :8
    ]  # Taking the first 8 characters for brevity

    # For HTML ids, we want to replace only the value inside the quotes
    if token_class == 'id' and matched_token.startswith('id='):
        return 'id="' + matched_token.split('"')[1] + '_' + unique_suffix + '"'

    # For JavaScript getElementById, we want to replace only the value inside the parentheses
    if token_class == 'id' and matched_token.startswith('document.getElementById('):
        return (
            'document.getElementById("'
            + matched_token.split('"')[1]
            + '_'
            + unique_suffix
            + '")'
        )

    # For CSS ids, we can append directly
    return matched_token + '_' + unique_suffix


# html_extractor = {"id": r'id="([^"]+)"'}

css_extractor = {'id': r'#([a-zA-Z][\w\-]*)'}

# js_extractor = {"id": r'document\.getElementById\("([^"]+)"\)'}

# Updated extractors
html_extractor = {'id': r'id\s*=\s*["\']([^"\']+?)["\']'}
js_extractor = {'id': r'document\.getElementById\s*\(\s*["\']([^"\']+?)["\']\s*\)'}


replace_html_ids = partial(
    replace_tokens,
    token_extractor=html_extractor,
    token_replacer=default_token_replacer,
)
replace_css_ids = partial(
    replace_tokens, token_extractor=css_extractor, token_replacer=default_token_replacer
)
replace_js_ids = partial(
    replace_tokens, token_extractor=js_extractor, token_replacer=default_token_replacer
)


# Determining the kind of string (html, css, or js)
# TODO: Make the default rules more robust


def _is_html_string(string):
    return '<div' in string or '<a' in string or '<span' in string


def _is_js_string(string):
    return (
        'document.getElementById' in string
        or 'function' in string
        or '.addEventListener' in string
    )


def _is_css_string(string):
    return '{' in string and '}' in string and (':' in string or ';' in string)


_dflt_string_kind_rules = tuple(
    {'html': _is_html_string, 'js': _is_js_string, 'css': _is_css_string}.items()
)


def _string_kind(string, string_kind_rules=_dflt_string_kind_rules):
    string_kind_rules = dict(string_kind_rules)
    for kind, rule in string_kind_rules.items():
        if rule(string):
            return kind


# TODO: Replacer logic is also a mapping. Perhaps could centralize with rules mapping
_dflt_replacer_for_kind = tuple(
    {'html': replace_html_ids, 'js': replace_js_ids, 'css': replace_css_ids}.items()
)


def replace_ids(
    string,
    kind: Optional[str] = None,
    *,
    replacer_for_kind=_dflt_replacer_for_kind,
    string_to_kind: Optional[Callable] = _string_kind
):
    kind = kind or string_to_kind(string)
    replacer_for_kind = dict(replacer_for_kind)
    replacer = replacer_for_kind.get(kind, None)
    if replacer is not None:
        return replacer(string)
    # if all else fails, just:
    return string, {}


# Tests ----------------------------------------------------------------
def test_replace_tokens():
    test_string = 'Hello, my id="name" is John.'
    extractor = {'id': r'id="([^"]+)"'}
    replacer = default_token_replacer
    new_string, replaced = replace_tokens(test_string, extractor, replacer)
    assert (
        new_string != test_string
    )  # The new string should be different due to the unique ID
    assert 'id="name_' in new_string  # The new ID should start with "name_"


def test_string_kind():
    assert _string_kind('<div class="test">Hello</div>') == 'html'
    assert _string_kind('#test { color: red; }') == 'css'
    assert _string_kind('function test() { alert("Hello"); }') == 'js'
    assert _string_kind('This is a random string.') == None


def test_replace_ids():
    html_test = '<div id="test">Hello</div>'
    css_test = '#test { color: red; }'
    js_test = 'document.getElementById("test").style.color = "red";'

    new_html, replaced_html = replace_ids(html_test)
    new_css, replaced_css = replace_ids(css_test)
    new_js, replaced_js = replace_ids(js_test)

    # Check if the replaced IDs have the unique suffix
    assert 'id="test_' in new_html
    assert '#test_' in new_css
    assert 'document.getElementById("test_' in new_js

    # Check the replaced dictionary
    old_html_id = list(replaced_html.keys())[0]
    old_css_id = list(replaced_css.keys())[0]
    old_js_id = list(replaced_js.keys())[0]

    assert old_html_id == 'id="test"'
    assert old_css_id == '#test'
    assert old_js_id == 'document.getElementById("test")'


## Run the tests
# test_replace_tokens()
# test_string_kind()
# test_replace_ids()

## print("All tests passed!")


# default_token_replacer('id="test"', 'id')
# 'id="test_<some_unique_suffix>"'



"""General utils"""

from functools import partial
import re

from dol import (
    TextFiles,
    wrap_kvs,
    filt_iter,
    invertible_maps,
    add_ipython_key_completions,
)

try:
    import importlib.resources

    _files = importlib.resources.files  # only valid in 3.9+
except AttributeError:
    import importlib_resources  # needs pip install

    _files = importlib_resources.files

files = _files('jy')
# data_dir = files / 'data'
# data_dir_path = str(data_dir)
js_dir = files / 'js'
js_dir_path = str(js_dir)


@add_ipython_key_completions
@wrap_kvs(key_of_id=lambda x: x[: -len('.js')], id_of_key=lambda x: x + '.js')
@filt_iter(filt=lambda x: x.endswith('.js'))
class JsFiles(TextFiles):
    """A store of js files"""


_replace_non_alphanumerics_by_underscore = partial(re.compile(r'\W').sub, '_')

# Note: js_files_as_attrs is not used in the module, but can be useful when working
# in a notebook, or console, where we might want the convenience of tab-completion of
# attributes
def js_files_as_attrs(rootdir):
    """
    Will make a JsFiles, but where the keys are available as attributes.
    To do so, any non alphanumerics of file name are replaced with underscore,
    and there can be no two files that collide with that key transformation!
    """
    from dol.sources import AttrContainer

    s = JsFiles(rootdir)
    key_for_id = {id_: _replace_non_alphanumerics_by_underscore(id_) for id_ in s}
    key_for_id, id_for_key = invertible_maps(key_for_id)
    return AttrContainer(
        **wrap_kvs(s, key_of_id=key_for_id.get, id_of_key=id_for_key.get)
    )


# Note: Could replace with js_files_as_attrs, but not sure if it's worth it
mk_js_files = JsFiles

js_files = mk_js_files(str(js_dir_path))



"""Tests for the bridge module"""

from jy.bridge import add_js_funcs

from jy.util import js_files

test01_js_code = js_files['test01']


def test_add_js_call_attributes_to_obj():
    """
    See test_js_parse.test_func_name_and_params_pairs for intermediate data of this test
    """
    js = add_js_funcs(test01_js_code)

    # js has two methods called bar and foo
    assert sorted([x for x in dir(js) if not x.startswith('_')]) == [
        'add_one',
        'bar',
        'foo',
        'obj',
        'prop',
        'with_arrow_func',
        'with_let',
    ]

    # they mirror the signatures of the underlying JS functions
    from inspect import signature

    assert str(signature(js.foo)) == "(a, b='hello', c=3)"
    assert str(signature(js.bar)) == "(green, eggs='food', and=True, ham=4)"

    # Calling this function returns a string
    # (the code to call the underlying JS function)
    assert js.foo(1, 'hi', 5) == 'foo(1, "hi", 5)'

    # Notice that you can use positional or keyword arguments
    # Also, notice that though "prop" is the name of js's attribute,
    # the function call string does indeed use the original full reference:
    # ``func.assigned.to.nested.prop``
    assert js.prop('up') == ('func.assigned.to.nested.prop("up")')

    # Notice that the python (signature) defaults are applied before translating to JS
    assert js.bar(42) == 'bar(42, "food", true, 4)'

    alt_js = add_js_funcs(test01_js_code, apply_defaults=False)
    # You can opt not to do this by specifying apply_defaults=False
    # This will result in only injecting those inputs you specify in the js call string,
    # which will have the effect of letting JS apply it's defaults, what ever they are
    alt_js = add_js_funcs(test01_js_code, apply_defaults=False)
    assert alt_js.bar(42) == 'bar(42)'



"""Tests for js_bridge"""

from jy.js_parse import func_name_and_params_pairs
from jy.util import js_files

test01_js_code = js_files['test01']


def test_func_name_and_params_pairs():
    pairs = list(func_name_and_params_pairs(test01_js_code))
    assert pairs == [
        (
            'foo',
            [
                {'name': 'a'},
                {'name': 'b', 'default': 'hello'},
                {'name': 'c', 'default': 3},
            ],
        ),
        (
            'bar',
            [
                {'name': 'green'},
                {'name': 'eggs', 'default': 'food'},
                {'name': 'and', 'default': True},
                {'name': 'ham', 'default': 4},
            ],
        ),
        ('add_one', [{'name': 'x'}]),
        ('with_let', [{'name': 'x'}]),
        ('with_arrow_func', [{'name': 'y'}, {'name': 'z', 'default': 1}]),
        # Note that the name here is dot-separated!
        ('func.assigned.to.nested.prop', [{'name': 'x'}]),
        ('obj', [{'name': 'exports'}]),
    ]


def test_replace_ids_in_code():
    from jy.js_parse import replace_ids_in_code

    # Test with HTML content
    html_code = '<div id="test">Hello</div>'
    new_html, replaced_html = replace_ids_in_code(html_code)
    assert 'id="test_' in new_html
    assert 'test' in replaced_html

    # Test with CSS content
    css_code = '#test { color: red; }'
    new_css, replaced_css = replace_ids_in_code(css_code)
    assert '#test_' in new_css
    assert 'test' in replaced_css

    # Test with JS content
    js_code = 'document.getElementById("test").style.color = "red";'
    new_js, replaced_js = replace_ids_in_code(js_code)
    assert 'document.getElementById("test_' in new_js
    assert 'test' in replaced_js

    # Test with mixed content
    mixed_code = '''
    <div id="test">Hello</div>
    <style>#test { color: red; }</style>
    <script>
    document.getElementById("test").style.color = "blue";
    </script>
    '''
    new_mixed, replaced_mixed = replace_ids_in_code(mixed_code)
    assert 'id="test_' in new_mixed
    assert '#test_' in new_mixed
    assert 'document.getElementById("test_' in new_mixed
    assert 'test' in replaced_mixed

    # Test with custom patterns
    custom_patterns = (r'id="custom_([^"]+)"',)
    custom_code = '<div id="custom_test">Hello</div>'
    new_custom, replaced_custom = replace_ids_in_code(custom_code, custom_patterns)
    assert 'id="custom_test_' in new_custom
    assert 'custom_test' in replaced_custom

    # Another test

    js_code = '''
    function sendToPython(){
        var data = document.querySelector("#myInput").value;
        var kernel = IPython.notebook.kernel;
        kernel.execute("data_from_js = '" + data + "'");
        kernel.execute("print('Received data from JS: ' + data_from_js)");
    }

    document.querySelector("#myButton").addEventListener("click", function(){
        sendToPython();
    });
    '''

    _js_code, _js_replacements = replace_ids_in_code(js_code)

    # Html (with original js_code embedded in it)
    HTML_code = '''
    <input type="text" id="myInput" placeholder="Enter some text">
    <button id="myButton">Submit</button>
    <script type="text/Javascript">{}</script>
    '''.format(
        js_code
    )

    _html_code, _html_replacements = replace_ids_in_code(HTML_code)

    assert (
        sorted(_js_replacements)
        == sorted(_html_replacements)
        == ['myButton', 'myInput']
    )

    for replacement in _html_replacements.values():
        assert replacement in _html_code
